/*
 * circle.c: Explicit Bresenham implementation of draw_circle() in
 * terms of draw_rectangle(), for the benefit of any platform which
 * doesn't have a sensible circle-drawing graphics primitive.
 */

#include <assert.h>
#include "palm-extras.h"

static void draw_circlex(frontend *fe, int cx, int cy, int radius,
                         int colour, int fill) PUZZ_SECT;
static void draw_circlex(frontend *fe, int cx, int cy, int radius,
                         int colour, int fill)
{
    /*
     * Bresenham's algorithm is used to compute a single octant of
     * a circle, from (say) the topmost point to one of the
     * 45-degree points. This is then replicated, by reflection and
     * rotation, into the other seven octants.
     * 
     * The way it works is that we're plotting y such that y^2 is
     * close to r^2-x^2, and we always keep track of the
     * _difference_ D between y^2 and r^2-x^2. So to start with,
     * we plot at (0,r), so that y=r, x=0 and D=0.
     * 
     * Next, we want to work out the correct y for x+1. Now (x+1)^2
     * equals x^2+2x+1, so the new value of (r^2-x^2) will be 2x+1
     * less than the old one - so D will go up by 2x+1. Now in this
     * octant, we know that we'll always be either decrementing y
     * by one, or leaving it alone; (y-1)^2 = y^2-2y+1, so _if_ we
     * decrement y then the new value of D will go down by 2y-1.
     * 
     * So we add 2x+1 to D, and then consider whether leaving it
     * alone or subtracting 2y-1 will make it end up closer to
     * zero. If the latter, we do so, and decrement y. Then we
     * increment x, plot the point, and repeat. The termination
     * condition comes when x>y, since then we've crossed over into
     * the next octant.
     * 
     * Since the algorithm is effectively rounding to nearest
     * rather than up or down, it doesn't suffer from little blobs
     * at the compass points or from excessively flat sides.
     */

    int x, y, D, D2;
    int ypolar, xpolar, ytropic;

    y = radius;
    x = 0;
    D = 0;

    ypolar = y;
    xpolar = 0;
    ytropic = 0;

    while (y >= x) {
        /*
         * Plot this point.
         */
        if (fill) {
            /*
             * We draw a filled circle as a sequence of horizontal
             * lines. So for _every_ point we plot, we can draw
             * horizontal lines for the part of the circle between
             * the tropics; but we only draw a line in the polar
             * region when we've finished extending it for a given
             * pixel.
             */
            palm_draw_rect(fe, cx-y, cy+x, 2*y+1, 1, colour);
            palm_draw_rect(fe, cx-y, cy-x, 2*y+1, 1, colour);
            ytropic = x;

            if (ypolar != y) {
                palm_draw_rect(fe, cx-xpolar, cy+ypolar, 2*xpolar+1, 1, colour);
                palm_draw_rect(fe, cx-xpolar, cy-ypolar, 2*xpolar+1, 1, colour);
                ypolar = y;
            }
            xpolar = x;
        } else {
          /*
           * For the outline circle, we simply plot the point in
           * all eight octants.
           */
          palm_draw_rect(fe, cx+x, cy+y, 1, 1, colour);
          palm_draw_rect(fe, cx+x, cy-y, 1, 1, colour);
          palm_draw_rect(fe, cx-x, cy+y, 1, 1, colour);
          palm_draw_rect(fe, cx-x, cy-y, 1, 1, colour);
          palm_draw_rect(fe, cx+y, cy+x, 1, 1, colour);
          palm_draw_rect(fe, cx+y, cy-x, 1, 1, colour);
          palm_draw_rect(fe, cx-y, cy+x, 1, 1, colour);
          palm_draw_rect(fe, cx-y, cy-x, 1, 1, colour);
        }

        D += 2*x + 1;
        D2 = D - (2*y - 1);
        if (abs(D2) < abs(D)) {
            D = D2;
            y--;
        }
        x++;
    }

    if (fill && ypolar > ytropic) {
        /*
         * Draw the last line in the polar region.
         */
        palm_draw_rect(fe, cx-xpolar, cy+ypolar, 2*xpolar+1, 1, colour);
        palm_draw_rect(fe, cx-xpolar, cy-ypolar, 2*xpolar+1, 1, colour);
    }
}

/* We have to do this separately, one after the other, because
 * otherwise the (delay-drawn) lines in the polar region overwrite
 * the points the outline half had previously placed there. */
void palm_draw_circle(void *handle, int cx, int cy, int radius,
                 int fillcolour, int outlinecolour)
{
    frontend *fe = (frontend *)handle;

    assert(outlinecolour >= 0);
    if (fillcolour >= 0)
        draw_circlex(fe, cx, cy, radius, fillcolour, 1);
    draw_circlex(fe, cx, cy, radius, outlinecolour, 0);
}
