/* fiveway.c
 *
 * Support for the various flavours of the 5way navigator.
 * There are two major flavours:
 * 1. The original Palm(One) scheme using vchrNavChange and the actual 5way key
 *    press or release in the keyCode field (Tungsten T, T2, T3, E, Zires)
 * 2. The HandSpring scheme using vchrRocker{Up,Down,Left,Right,Center} in
 *    the chr field and either the key press/release in the keyCode field
 *    (same as the Palm(One) scheme) or the chr field duplicated in the 
 *    keyCode field (Tungsten T5, E2, LifeDrive, Treo 600/650). */

#include "palm-extras.h"

#define vchrNavChange	(vchrPalmMin + 3)

#define navBitUp	0x0001      /* mask for the five way UP button. */
#define navBitDown	0x0002      /* mask for the five way DOWN button. */
#define navBitLeft	0x0004      /* mask for the five way LEFT button. */
#define navBitRight	0x0008      /* mask for the five way RIGHT button. */
#define navBitSelect	0x0010      /* mask for the five way SELECT button. */
#define navBitsAll	0x001F      /* mask for all five way buttons. */

#define navChangeUp	0x0100      /* mask for the five way UP button state change. */
#define navChangeDown	0x0200      /* mask for the five way DOWN button state change. */
#define navChangeLeft	0x0400      /* mask for the five way LEFT button state change. */
#define navChangeRight	0x0800      /* mask for the five way RIGHT button state change. */
#define navChangeSelect	0x1000      /* mask for the five way SELECT button state change. */
#define navChangeBitsAll 0x1F00     /* mask for all five way buttons state change. */

static Boolean is_5way_palm_event(EventPtr eventP) CUSTOM_SECT;
static Boolean is_5way_palm_event(EventPtr eventP)
{
  Boolean ret = false;

  if (((eventP->data.keyDown.chr == vchrPageUp)   ||
       (eventP->data.keyDown.chr == vchrPageDown) ||
       (eventP->data.keyDown.chr == vchrNavChange) ) &&
      ((eventP->data.keyDown.keyCode & (navBitsAll|navChangeBitsAll)) != 0))
    ret = true;
  debug(("is_5way_palm_event returns %s", ret ? "true" : "false"));
  return ret;
}

static Boolean is_rocker_key(UInt16 modifiers, WChar chr) CUSTOM_SECT;
static Boolean is_rocker_key(UInt16 modifiers, WChar chr)
{
  Boolean ret = false;
  if ((modifiers & commandKeyMask) &&
      (chr >= vchrRockerUp) &&
      (chr <= vchrRockerCenter))
    ret = true;
  debug(("is_rocker_key returns %s", ret ? "true" : "false"));
  return ret;
}

static Boolean is_5way_event(EventPtr eventP) CUSTOM_SECT;
static Boolean is_5way_event(EventPtr eventP)
{
  Boolean ret = false;
  if (is_5way_palm_event(eventP) ||
      is_rocker_key(eventP->data.keyDown.modifiers, eventP->data.keyDown.chr) ||
      (eventP->data.keyDown.chr == vchrHardRockerCenter))
    ret = true;
  debug(("is_5way_event returns %s", ret ? "true" : "false"));
  return ret;
}

static Boolean center_hs_pressed(EventPtr eventP) CUSTOM_SECT;
static Boolean center_hs_pressed(EventPtr eventP)
{
  Boolean ret = false;
  if ((eventP->data.keyDown.modifiers & commandKeyMask) &&
      ((eventP->data.keyDown.chr == vchrRockerCenter) ||
       ((eventP->data.keyDown.chr == vchrHardRockerCenter) &&
        ((eventP->data.keyDown.keyCode & (navBitsAll|navChangeBitsAll)) == navChangeSelect))))
    ret = true;
  debug(("center_hs_pressed returns %s", ret ? "true" : "false"));
  return ret;
}

static Boolean center_palm_pressed(EventPtr eventP) CUSTOM_SECT;
static Boolean center_palm_pressed(EventPtr eventP)
{
  Boolean ret = false;
  if (((eventP->data.keyDown.modifiers & autoRepeatKeyMask) == 0) &&
      (eventP->data.keyDown.chr == vchrNavChange) &&
      ((eventP->data.keyDown.keyCode & (navBitsAll|navChangeBitsAll)) == navChangeSelect))
    ret = true;
  debug(("center_palm_pressed returns %s", ret ? "true" : "false"));
  return ret;
}

static Boolean center_pressed(EventPtr eventP) CUSTOM_SECT;
static Boolean center_pressed(EventPtr eventP)
{
  Boolean ret = false;
  if (is_5way_event(eventP)) {
    ret = (center_palm_pressed(eventP) || center_hs_pressed(eventP));
  }
  debug(("center_pressed returns %s", ret ? "true" : "false"));
  return ret;
}

static Boolean dir_hs_pressed(EventPtr eventP, Direction *dir) CUSTOM_SECT;
static Boolean dir_hs_pressed(EventPtr eventP, Direction *dir)
{
  Boolean ret = false;
  UInt16 navdir, navchangedir;
  UInt32 osver, osver549;

  /* We are checking for PalmOS 5.4.9 in order to make the fiveway navigator
   * work for the Palm TX and Z22. They changed the fiveway keycodes again ...
   * Moreover the keycodes emitted by the real device are not the same as
   * the ones emitted by the simulator. Sigh ... */
  FtrGet(sysFtrCreator, sysFtrNumROMVersion, &osver);
  osver549 = sysMakeROMVersion(5,4,9,0,0);

  if (!(eventP->data.keyDown.modifiers & commandKeyMask)) {
    ret = false;
  }
  else if (eventP->data.keyDown.chr >= vchrRockerUp && eventP->data.keyDown.chr <= vchrRockerRight) {
    *dir = eventP->data.keyDown.chr - vchrRockerUp;
    navdir = navBitUp << *dir;
    navchangedir = navChangeUp << *dir;
    if (((eventP->data.keyDown.keyCode & (navBitsAll|navChangeBitsAll)) == navdir) ||
        ((osver >= osver549) && ((eventP->data.keyDown.keyCode & (navBitsAll|navChangeBitsAll)) == navchangedir)) ||
        (eventP->data.keyDown.chr == eventP->data.keyDown.keyCode))
      ret = true;
  }
  else if ((eventP->data.keyDown.chr == vchrPageUp) ||
           (eventP->data.keyDown.chr == vchrPageDown)) {
    *dir = eventP->data.keyDown.chr - vchrPageUp;
    navdir = navBitUp << *dir;
    navchangedir = navChangeUp << *dir;
    if (((eventP->data.keyDown.keyCode & (navBitsAll|navChangeBitsAll)) == navdir) ||
        ((osver >= osver549) && ((eventP->data.keyDown.keyCode & (navBitsAll|navChangeBitsAll)) == navchangedir)) ||
        (eventP->data.keyDown.chr == eventP->data.keyDown.keyCode))
      ret = true;
  }
  debug(("dir_hs_pressed returns %s, dir = %d", ret ? "true" : "false", ret ? *dir : -1));
  return ret;
}

static Boolean dir_palm_pressed(EventPtr eventP, Direction *dir) CUSTOM_SECT;
static Boolean dir_palm_pressed(EventPtr eventP, Direction *dir)
{
  Boolean ret = false;
  UInt16 nav_and_change_bits, nav_bits, change_bits;

  if (eventP->data.keyDown.modifiers & autoRepeatKeyMask) {
    /* For autorepeat only navBitxx is set, navChangexx is not set */
    nav_and_change_bits = eventP->data.keyDown.keyCode & (navBitsAll|navChangeBitsAll);
    if (nav_and_change_bits >= navBitUp && nav_and_change_bits <= navBitRight) {
      switch (nav_and_change_bits) {
      case navBitUp:
        *dir = Up;
        break;
      case navBitDown:
        *dir = Down;
        break;
      case navBitLeft:
        *dir = Left;
        break;
      case navBitRight:
        *dir = Right;
        break;
      }
      ret = true;
    }
  }
  else {
    /* For a single key press, both navBitxx and corresponding navChangexx are set */
    nav_bits = eventP->data.keyDown.keyCode & navBitsAll;
    change_bits = (eventP->data.keyDown.keyCode & navChangeBitsAll) >> 8;
    if (nav_bits == change_bits && nav_bits >= navBitUp && nav_bits <= navBitRight) {
      switch (nav_bits) {
      case navBitUp:
        *dir = Up;
        break;
      case navBitDown:
        *dir = Down;
        break;
      case navBitLeft:
        *dir = Left;
        break;
      case navBitRight:
        *dir = Right;
        break;
      }
      ret = true;
    }
  }
  debug(("dir_palm_pressed returns %s, dir = %d", ret ? "true" : "false", ret ? *dir : -1));
  return ret;
}

static Boolean dir_pressed(EventPtr eventP, Direction *dir) CUSTOM_SECT;
static Boolean dir_pressed(EventPtr eventP, Direction *dir)
{
  Boolean ret = false;
  if (is_5way_event(eventP)) {
    ret = (dir_palm_pressed(eventP, dir) || dir_hs_pressed(eventP, dir));
  }
  else if (eventP->data.keyDown.chr == vchrPageUp) {
    *dir = Up;
    ret = true;
  }
  else if (eventP->data.keyDown.chr == vchrPageDown) {
    *dir = Down;
    ret = true;
  }
  debug(("dir_pressed returns %s, dir = %d", ret ? "true" : "false", ret ? *dir : -1));
  return ret;
}

Boolean fiveway_key(EventPtr eventP, Direction *dir)
{
  Boolean ret = false;
  if (center_pressed(eventP)) {
    *dir = Center;
    ret = true;
  }
  else
    ret = dir_pressed(eventP, dir);
  debug(("5way_key returns %s, dir = %d", ret ? "true" : "false", ret ? *dir : -1));
  return ret;
}

