#include "puzzles.h"
#include <MemGlue.h>

#ifdef DEBUG_MALLOC
#define mdebug(x) (debug_printf x)
#else
#define mdebug(x)
#endif


/* This is all a bit scungy currently. The palm docs advise against
 * using MemPtr*, and say 'use MemHandle* instead', which is dynamic
 * and needs locking and all sorts of other stuff.
 *
 * Possible improvements:
 *
   * Use MemHandleFoo instead of MemPtrFoo (but what about locking?)
   * Use MemGlue... which cope with larger blocks.
 *
 * ...and I hope it's safe to cast void* to MemPtr etc.
 */
#ifdef DEBUG_MALLOC
void *smalloc_d(size_t size, const char *file, int line)
#else
void *smalloc(size_t size)
#endif
{
  MemPtr p;

  if (size == 0) size = 1; /* palmos returns NULL from malloc(0). */
  p = MemGluePtrNew(size);

  if (!p)
    mdebug(("%s:%d alloc %ld (0x%lx) --> FAILED", file, line, size, size));
  else
    mdebug(("%s:%d alloc %ld (0x%lx) --> 0x%lx", file, line, size, size, p));

  if (!p) fatal_malloc(size);
  return p;
}

#ifdef DEBUG_MALLOC
void sfree_d(void *p, const char *file, int line)
#else
void sfree(void *p)
#endif
{
  if (!p) {
    mdebug(("%s:%d free 0", file, line));
    return;
  }
  mdebug(("%s:%d free 0x%lx (size 0x%lx)", file, line, p, MemPtrSize(p)));
  MemPtrFree(p);
}

#ifdef DEBUG_MALLOC
void *srealloc_d(void *p, size_t size, const char *file, int line)
#else
void *srealloc(void *p, size_t size)
#endif
{
  MemPtr q = NULL;

  if (!p) {
    mdebug(("%s:%d realloc from 0 == malloc...", file, line));
#ifdef DEBUG_MALLOC
    return smalloc_d(size, file, line);
#else
    return smalloc(size);
#endif
  }
  if (!size) {
    mdebug(("%s:%d realloc to size 0 == free...", file, line));
#ifdef DEBUG_MALLOC
    sfree_d(p, file, line);
#else
    sfree(p);
#endif
  }

  /* we now know we're doing a real realloc. */
  if (MemPtrResize(p, size) == 0) {
    q = p;
  } else {
    q = MemGluePtrNew(size);
    if (q) {
      MemMove(q, p, MemPtrSize (p));
      MemPtrFree(p);
    }
  }
  if (!q) {
    mdebug(("%s:%d realloc 0x%lx to %ld (0x%lx) --> FAILED",
            file, line, p, size, size));
    MemPtrFree(p); /* XXX debatable */
    fatal_malloc(size);
  }
  mdebug(("%s:%d realloc 0x%lx to %ld (0x%lx) --> 0x%lx",
          file, line, p, size, size, q));
  return q;
}

#ifdef DEBUG_MALLOC
char *dupstr_d(const char *s, const char *file, int line)
#else
char *dupstr(const char *s)
#endif
{
  char *r;

  mdebug(("%s:%d dupstr '%s' at 0x%lx, about to alloc...",
         file, line, s, s));

#ifdef DEBUG_MALLOC
  r = smalloc_d(1+StrLen(s), file, line);
#else
  r = smalloc(1+StrLen(s));
#endif

  StrCopy(r,s);
  return r;
}
