#! /usr/bin/env python

import os
import sys
import re
import glob
import time

def die(s):
  print >>sys.stderr, s
  sys.exit(1)

def mksect(fstem):
  return "g_%s" % (fstem[0:6])

def mtime(file):
  return os.stat(file)[8]


if len(sys.argv) < 2: usage()

unwanted = [ "nullgame" ]

class Sectionizer:
  def __init__(self):
    self.sections = []   # section names
    self.srcsect = {}    # fname --> (s1, s2, fns for s2)
    self.buildfiles = {} # dest file --> src file
    self.puzzles = {}    # puzzle names, with dependencies
    self.srcdirs = {}    # directories to scan: NAME --> directory
    self.srcs = {}       # C files in above: NAME --> [stem1, stem2, ...]
    self.stems = []      # Stems of all files we've scanned.
    self.builddir = ""

  def set_builddir(self, builddir):
    self.builddir = builddir
    if not os.path.exists(self.builddir):
      os.mkdir(self.builddir)

  def puzz_depends(self, puzzle, rname):
    # this simplifies the parsing, especially for complex .R files
    # (e.g. keen.R). 
    rfile = open(rname)
    variables = {}
    depends = set()
    for l in rfile:
      lmo = re.match('^([A-Z_]+)?\s*=(.+)$', l)
      if lmo:
        variables[lmo.group(1)] = lmo.group(2).strip().split(' ')
        depends = depends | set(variables[lmo.group(1)])
    if depends != set():
      if puzzle in depends:
        depends.remove(puzzle)
      depends = depends - (set(variables.keys()))
      print "(depends: %s)" % (depends),
    return depends

  def add_dir(self, name, dirname):
    self.srcdirs[name] = dirname
    self.srcs[name] = []
    print "Scanning %s files in %s: " % (name, dirname)

    chfiles = glob.glob(os.path.join(dirname, "*.[ch]"))
    for srcf in chfiles:
      if srcf == 'puzzles.h': continue # we deal with this separately.
      dstf = os.path.join(self.builddir, os.path.basename(srcf))
      if dstf in self.buildfiles:
        die("%s duplicated at %s", self.buildfiles[dstf], srcf)
      self.buildfiles[dstf] = srcf

      (stem, ext) = os.path.splitext(srcf)
      puzz = os.path.basename(stem)
      self.stems.append(puzz)
      if ext == '.c':
        self.srcs[name].append(puzz)

      rfile = stem + '.R'
      print "  " + puzz + ext,
      if not puzz in unwanted and os.path.exists(rfile):
        puzzs = mksect(puzz)
        if puzz not in self.puzzles:
          self.puzzles[puzz] = self.puzz_depends(puzz, rfile)
        if puzz not in self.srcsect:
          self.srcsect[puzz] = (puzzs, '', [])
        if puzzs not in self.sections:
          self.sections.append(puzzs)
      print

  def set_sections(self, conffile):
    print "Reading section configuration from %s..." % (conffile)
    sconf = open(conffile)
    for l in sconf:
      l = l.strip()
      if l == '' or l.startswith('#'): continue
      
      attrs = re.sub(r'\s+', ' ', l).split(' ')
      
      filestem = attrs[0]
      if filestem in unwanted: continue
    
      sects = attrs[1]
      if sects.count('/') == 0:
        sect = mksect(sects)
        sect2 = ""
      else:
        (sect, sect2) = map(mksect, sects.split('/'))
    
      if len(attrs) == 3:
        fns = attrs[2].split(',')
      else:
        fns = []
    
      if filestem == '_section':
        if sect not in self.sections: self.sections.append(sect)
      elif filestem in self.stems:
        self.srcsect[filestem] = (sect, sect2, fns)
        if sect not in self.sections: self.sections.append(sect)
        if sect2 not in self.sections: self.sections.append(sect2)

  def write_sect(self, fname):
    if os.path.exists(fname) and mtime(fname) >= mtime(self.buildfiles[fname]):
      # only recreate the dest if it doesn't exist,
      # or is older than the source.
      return 0
    
    (name, ext) = os.path.splitext(os.path.basename(fname))
    fnsects = {}
    
    def fixup(mo):
      (decl, fnname, post) = mo.group(1,2,3)
    
      # if our declaration doesn't have preprocessor guff,
      # put it on one line. If it does, we lose (and will smash __LINE__).
      decl_oneline = decl
      if decl.count('#') == 0:
        decl_oneline = re.sub(r'\s+', ' ', decl_oneline)
    
      dsect = self.srcsect[name][0]
      for fn in self.srcsect[name][2]:
        if decl_oneline.count(fn) > 0: dsect = self.srcsect[name][1]
    
      ret = "%s __attribute__ ((section (\"%s\")));" % (decl, dsect)
      if post != ';':
        ret = ret + " %s\n{" % (decl_oneline)
        fnsects[fnname] = dsect
      return ret
    
    ip = open(self.buildfiles[fname])
    lines = ip.read() # whole file at once.
    ip.close()
    
    if name in self.srcsect:
      if ext == '.h':
        lines = re.sub(r'(?m)^(\w+[^({=]+?([\w_]+)\([^)]+\))(\s+{|;)', fixup, lines)
      else:
        lines = re.sub(r'(?m)^(static[^({=]+?([\w_]+)\([^)]+\))(\s+{|;)', fixup, lines)
    
    print "Writing %s (%d modified declarations)..." % (fname, len(fnsects))
    op = open(fname, 'w')
    op.write(lines)
    print >>op, "\n"
    print >>op, "/***********************************************";
    print >>op, " * PalmOS section information added by mksect.py";
    print >>op, " * " + time.asctime(time.gmtime(time.time()))
    print >>op, " *"
    for fnname in fnsects:
      print >>op, " *   %s:\t %s" % (fnsects[fnname], fnname)
    print >>op, " ***********************************************/";
    op.close()
    return 1

  def write_sects(self):
    nwritten = 0
    for f in self.buildfiles:
      nwritten += self.write_sect(f)

  # TODO: allow this to take a puzzle, and write a def file for
  # a .prc including just this puzzle (and the rest).
  def write_def(self, deffile):
    pdef = open(os.path.join(self.builddir, deffile), 'w')
    print >>pdef, "application { \"Puzzles\" SGTp stack=16384 }"
    print >>pdef, "multiple code { ",
    for sect in self.sections:
      print >>pdef, sect + ' ',
    print >>pdef, "}"
    pdef.close()

  def write_list(self, listfile, rcpfile, ll=None):
    if ll is None:
      ll = self.puzzles.keys()
    ll.sort()

    lfile = open(os.path.join(self.builddir, listfile), 'w')
    print >>lfile, "/* THIS FILE IS AUTO-GENERATED! */"
    print >>lfile, "/* These are in the same order as docs.rcp. */\n"
    print >>lfile, "#include \"puzzles.h\""
    print >>lfile, ""
    for puzz in ll:
      print >>lfile, "extern const game %s;" % (puzz)
    print >>lfile, ""
    print >>lfile, "CONST_GLOBAL game *gamelist[] = {"
    for puzz in ll:
      print >>lfile, "    (CONST_GLOBAL game*)&%s," % (puzz)
    print >>lfile, "};"
    print >>lfile, ""
    print >>lfile, "CONST_GLOBAL int gamecount = %d;" % (len(ll))
    lfile.close()

    rcp = open(os.path.join(self.builddir, rcpfile), 'w')
    print >>rcp, "/* THIS FILE IS AUTO-GENERATED */"
    print >>rcp, "/* These are in the same order as list-palm.c. */\n"
    hcount = 0
    for puzz in ll:
      print >>rcp, "STRING ID hlpHelp%d\tFILE \"../docs/%s.txt\"" % (hcount, puzz)
      hcount += 1
    rcp.close()

  def write_makefrag(self, fragfile):
    makefrag = open(os.path.join(self.builddir, fragfile), 'w')
    print >>makefrag, "PUZZLES= " + ' '.join(self.puzzles.keys())
    for name in self.srcs:
      print >>makefrag, name + "SRC= " + ' '.join(self.srcs[name])
    helpers = {}
    for puzz in self.puzzles:
      print >>makefrag, "HELPERS_%s= %s" % (puzz, ' '.join(self.puzzles[puzz]))
      for dep in self.puzzles[puzz]:
        helpers[dep] = 1
    print >>makefrag, "HELPERS= " + ' '.join(helpers.keys())
    makefrag.close()

def usage():
  die("Usage: mksect.py BUILDDIR NAME=DIR [NAME=DIR ...]")

if len(sys.argv) < 2: usage()

ss = Sectionizer()
ss.set_builddir(sys.argv[1])

for arg in sys.argv[2:]:
  (name, dirname) = arg.split('=')
  ss.add_dir(name, dirname)

ss.set_sections('sections.conf')

ss.write_sects()

print "Updating auto-generated files ..."
ss.write_def('puzzles.def')
ss.write_list('list-palm.c', 'docs.rcp')
ss.write_makefrag('mksect.make')

sys.exit(0)
