#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "puzzles.h"
#include "palm-extras.h"
#include "palmrcp.h"

/* --- Constrants, versioning, etc. --- */

/* This is the version of prefs we're currently saving. */
#define PREFS_VER 2

#define PREFS_STD_ID            0       /* Standard prefs struct */
#define PREFS_SEED_ID           1       /* v1 prefs stored game seed */
#define PREFS_SERIALISED_ID     2       /* v2 prefs serialised game */
#define PREFS_OPTIONS_ID        3       /* options struct */

#define PREFS_PRESET_ORIGIN     0x100   /* first stored preset */

/* --- Preference-type utility functions --- */

typedef union _hash {
  unsigned char bytes[20];
  long words[5];
} Hash;

typedef struct _prefix_hash {
  int v;
  Hash hash;
  unsigned char ptr[1];
} prefix_hash;

static int game2hash(int gameno, Hash *hash_r) GEN_SECT(saves);
static int game2hash(int gameno, Hash *hash_r)
{
  const char *gname;

  if (gameno >= gamecount) return (-1);
  gname = gamelist[gameno]->name;

  SHA_Simple((void*)gname, strlen(gname), hash_r->bytes);
  return (0);
}

static int hash2game(Hash *hash, int *gameno_r) GEN_SECT(saves);
static int hash2game(Hash *hash, int *gameno_r)
{
  Hash hashcmp;
  int i, ret;

  for (i = 0; i < gamecount; i++) {
    ret = game2hash(i, &hashcmp);
    assert(!ret);
    if (memcmp(hashcmp.bytes, hash->bytes, sizeof(Hash)) == 0) {
      *gameno_r = i;
      return (0);
    }
  }
  return (-1);
}

static UInt32 GetCreatorID(void) GEN_SECT(saves);
static UInt32 GetCreatorID(void)
{
  static UInt32 nCreatorID = 0;

  if (nCreatorID == 0) {
    UInt16 nCard;
    LocalID LocalDB;

    ErrFatalDisplayIf(SysCurAppDatabase(&nCard, &LocalDB),
                      "Could not get current app database.");
    ErrFatalDisplayIf(DmDatabaseInfo(nCard, LocalDB, 0, 0, 0, 0,
                                     0, 0, 0, 0, 0, 0, &nCreatorID),
                      "Could not get app database info, looking for creator ID");
#ifdef DEBUGGING
    {
      char *cid_c = (char *)&nCreatorID; /* scunge for debugging */
      debug(("Read creator ID 0x%lx ('%c%c%c%c')",
            nCreatorID, cid_c[0], cid_c[1], cid_c[2], cid_c[3]));
    }
#endif
  }

  return nCreatorID;
}

/* This sets anything that prefs might alter to sensible defaults,
 * in case we don't read prefs (for whatever reason). */
static void SetDefaults(void) GEN_SECT(saves);
static void SetDefaults(void)
{
  fe_global.currgame = 0;
}

static void *easy_read_pref(UInt32 cid, UInt16 id, Boolean backup, Int16 *sz_r) GEN_SECT(saves);
static void *easy_read_pref(UInt32 cid, UInt16 id, Boolean backup, Int16 *sz_r)
{
  Int16 sz = 0, pVer;
  void *read = NULL;

  pVer = PrefGetAppPreferences(cid, id, NULL, &sz, backup);
  if (pVer != noPreferenceFound) {
    read = smalloc(sz);
    PrefGetAppPreferences(cid, id, read, &sz, backup);
  } else
    sz = 0;
  if (sz_r) *sz_r = sz;
  return read;
}

/* --- Definitions of pref structs --- */

struct _pref1 {
  int lastgame, lastpreset;
  unsigned int flags;
};

struct _pref2 {
  Hash lastgame;
  unsigned int flags;   /* uses same PF_* as v1 prefs. */
  int ngameseeds;       /* the no. of games we stored seeds for. */
};

struct _prefs {
  int ver;
  union {
    struct _pref1 one;
    struct _pref2 two;
  } prefs;
  struct _options options;
};

struct _prefs savedprefs;
struct _options *options = &(savedprefs.options);

/* --- Old version 1 preferences --- */

#define PF_LOCKED 1

static int ReadPrefs_one(UInt32 cid, struct _pref1 *pref, UInt16 sz) GEN_SECT(saves);
static int ReadPrefs_one(UInt32 cid, struct _pref1 *pref, UInt16 sz)
{
  char *seed = NULL;

  assert(sz == sizeof(struct _pref1));
  memset(pref, 0, sz);
  PrefGetAppPreferences(cid, PREFS_STD_ID, pref, &sz, true);

  seed = easy_read_pref(cid, PREFS_SEED_ID, true, &sz);

  if (pref->lastgame < gamecount) {
    fe_global.currgame = pref->lastgame;
    /*fe_game(&fe_global)->preset = pref->lastpreset;*/
    /* XXX If we care, we should call midend_fetch_preset and then
     * midend_set_params with pref->lastpreset. */
    if (seed) midend_game_id(fe_me(&fe_global), seed);
  }
  if (pref->flags & PF_LOCKED)
    fe_global.flags |= FF_GAMELOCKED;

  if (seed) sfree(seed);

  return (0);
}

/*static void WritePrefs_one(UInt32 cid) GEN_SECT(saves);
static void WritePrefs_one(UInt32 cid)
{
  struct _pref1 sp;
  config_item *cfg;
  char *unused;

  sp.lastgame = fe_global.currgame;
  sp.flags = 0;
  if (fe_global.flags & FF_GAMELOCKED) sp.flags |= PF_LOCKED;

  PrefSetAppPreferences(cid, PREFS_STD_ID, 1,
                        &sp, sizeof(struct _pref1), true);
  debug(("WritePrefs: stored preferences version %d size %d",
         PREFS_VER, sizeof(struct _pref1)));

  cfg = midend_get_config(fe_me(&fe_global), CFG_DESC, &unused);
  if (cfg) {
    int slen = strlen(cfg[0].sval)+1;
    PrefSetAppPreferences(cid, PREFS_SEED_ID, 1,
                          cfg[0].sval, slen, true);
    debug(("WritePrefs: stored seed '%s' size %d in preferences",
           cfg[0].sval, slen));
    sfree(unused);
    free_cfg(cfg);
  }
}*/

/* --- Version 2 preferences --- */

static void ReadPrefs_two_presets(UInt32 cid, int nstored) GEN_SECT(saves);
static void ReadPrefs_two_presets(UInt32 cid, int nstored)
{
  int i, game;
  prefix_hash *stored;
  Int16 sz;

  for (i = 0; i < nstored; i++) {
    stored = easy_read_pref(cid, PREFS_PRESET_ORIGIN + i, false, &sz);
    if (!stored || sz < sizeof(prefix_hash)) {
      debug(("preset %d: stored pref too small (or absent).", i));
      continue;
    }
    if (stored->v > 1) {
      debug(("preset %d: version not understood.", i));
      goto next;
    }
    if (hash2game(&stored->hash, &game)) {
      debug(("preset %d: for unknown game", i));
      goto next;
    }
    debug(("Loading game seed '%s' into %s (game %d).",
           (char *)stored->ptr, gamelist[game]->name, game));
    midend_game_id(fe_global.pergame[game].me, (char *)stored->ptr);
next:
    sfree(stored);
  }
}

static int WritePrefs_two_presets(UInt32 cid, int nprevstored) GEN_SECT(saves);
static int WritePrefs_two_presets(UInt32 cid, int nprevstored)
{
  int i, sz;
  prefix_hash *stored;
  config_item *cfg;
  char *unused;

  for (i = 0; i < gamecount; i++) {
    cfg = midend_get_config(fe_global.pergame[i].me, CFG_SEED, &unused);
    if (!cfg) {
      debug(("Game '%s' (%d) has no config to store.",
             gamelist[i]->name, i));
      continue;
    }
    sz = sizeof(prefix_hash)+strlen(cfg[0].sval)+1;

    stored = smalloc(sz);
    stored->v = 1;
    assert(!game2hash(i, &stored->hash));
    memcpy(stored->ptr, cfg[0].sval, strlen(cfg[0].sval)+1);

    debug(("Saving game seed '%s' from %s (game %d).",
           (char *)stored->ptr, gamelist[i]->name, i));
    PrefSetAppPreferences(cid, PREFS_PRESET_ORIGIN + i, 1,
                          stored, sz, false);

    sfree(stored);
    sfree(unused);
    free_cfg(cfg);
  }
  /* Delete any excess prefs we might have stored last time. */
  if (gamecount < nprevstored) {
    for (i = gamecount; i < nprevstored; i++) {
      PrefSetAppPreferences(cid, PREFS_PRESET_ORIGIN + i, 1,
                          NULL, 0, false);
    }
  }
  return gamecount;
}

static int ReadPrefs_two(UInt32 cid, struct _pref2 *pref, UInt16 sz) GEN_SECT(saves);
static int ReadPrefs_two(UInt32 cid, struct _pref2 *pref, UInt16 sz)
{
  int game;
  unsigned char *serialised = NULL;
  char *err;

  debug(("Reading v2 prefs."));

  assert(sz == sizeof(struct _pref2));
  memset(pref, 0, sz);
  PrefGetAppPreferences(cid, PREFS_STD_ID, pref, &sz, true);

  ReadPrefs_two_presets(cid, pref->ngameseeds);

  serialised = easy_read_pref(cid, PREFS_SERIALISED_ID, true, &sz);

  if (!hash2game(&pref->lastgame, &game)) {
    debug(("v2 prefs saved game no. %d", game));
    fe_global.currgame = game;
    if (serialised) {
      debug(("Read serialised game."));
      //midend_new_game(fe_me(&fe_global));
      err = midend_im_deserialise(fe_me(&fe_global), serialised, sz);
      if (err)
        FrmCustomAlert(validationError, err, NULL, NULL);
      else {
        //midend_force_redraw(fe_me(&fe_global));
        fe_global.flags |= FF_NONEWGAME;
      }
    }
  }
  if (pref->flags & PF_LOCKED)
    fe_global.flags |= FF_GAMELOCKED;

  if (serialised) sfree(serialised);

  return (0);
}

static void WritePrefs_two(UInt32 cid) GEN_SECT(saves);
static void WritePrefs_two(UInt32 cid)
{
  struct _pref2 sp;
  unsigned char *serialised = NULL;
  size_t sz;

  memset(&sp, 0, sizeof(struct _pref2));

  assert(!game2hash(fe_global.currgame, &sp.lastgame));
  if (fe_global.flags & FF_GAMELOCKED) sp.flags |= PF_LOCKED;

  sp.ngameseeds = WritePrefs_two_presets(cid, savedprefs.prefs.two.ngameseeds);

  midend_im_serialise(fe_me(&fe_global), &serialised, &sz);
  if (serialised) {
    debug(("Serialised game."));
    PrefSetAppPreferences(cid, PREFS_SERIALISED_ID, 2,
                          serialised, sz, true);
    sfree(serialised);
  } else {
    debug(("Unable to serialise."));
  }

  PrefSetAppPreferences(cid, PREFS_STD_ID, 2,
                        &sp, sizeof(struct _pref2), true);

  /* delete old struct */
  PrefSetAppPreferences(cid, PREFS_SEED_ID, 1, NULL, 0, true);
}

/* --- Options stuff --- */

static void DefaultOptions(struct _options *opt) GEN_SECT(saves);
static void DefaultOptions(struct _options *opt)
{
    opt->rclick_millis = 250;
    opt->rclick_offset = 3;
}

char *CheckOptions(struct _options *opt)
{
    char *ret = NULL;

    if (opt->rclick_millis < 10) {
        opt->rclick_millis = 10;
        ret = "Right-click delay must be > 10 ms.";
    } else if (opt->rclick_millis > 10000) {
        opt->rclick_millis = 10000;
        ret = "Right-click delay should be < 10s.";
    }
    if (opt->rclick_offset < 0) {
        opt->rclick_offset = 0;
        ret = "Right-click drag offset must be >= 0.";
    } else if (opt->rclick_offset > 160) {
        opt->rclick_offset = 160;
        ret = "Right-click drag offset must be <= 160 pixels.";
    }
    return ret;
}

static void ReadOptions(UInt32 cid, struct _options *opt) GEN_SECT(saves);
static void ReadOptions(UInt32 cid, struct _options *opt)
{
  UInt16 pSize = 0;

  PrefGetAppPreferences(cid, PREFS_OPTIONS_ID, NULL, &pSize, true);

  memset(opt, 0, sizeof(struct _options));
  DefaultOptions(opt);

  if (pSize > sizeof(struct _options)) {
      FrmAlert(prefsCorrupt);
      return;
  }
  PrefGetAppPreferences(cid, PREFS_OPTIONS_ID, opt, &pSize, true);
  CheckOptions(opt); /* ignore return code, we just want it to fix it up. */
}

static void WriteOptions(UInt32 cid) GEN_SECT(saves);
static void WriteOptions(UInt32 cid)
{
  PrefSetAppPreferences(cid, PREFS_OPTIONS_ID, 1,
                        options, sizeof(struct _options), true);
}

/* --- General prefs stuff --- */

int ReadPrefs(void)
{
  UInt16 pSize;
  Int16 pVer;
  UInt32 cid = GetCreatorID();

  SetDefaults();

  memset(&savedprefs, 0, sizeof(struct _prefs));

  ReadOptions(cid, options);

  pSize = 0;
  pVer = PrefGetAppPreferences(cid, PREFS_STD_ID, NULL, &pSize, true);
  debug(("ReadPrefs: found prefs v%d, size %d bytes.", pVer, pSize));
  switch (pVer) {
  case noPreferenceFound:
    break;

  case 1:
    if (pSize != sizeof(struct _pref1))
      goto corrupt;
    savedprefs.ver = 1;
    ReadPrefs_one(cid, &(savedprefs.prefs.one), pSize);
    break;

  case 2:
    if (pSize != sizeof(struct _pref2))
      goto corrupt;
    savedprefs.ver = 2;
    ReadPrefs_two(cid, &(savedprefs.prefs.two), pSize);
    break;

  default:
    goto corrupt;
  }
  return 0;

corrupt:
  FrmAlert(prefsCorrupt);
  return (-1);
}

/* just write a v1 prefs struct for now. */
void WritePrefs(void)
{
  UInt32 cid = GetCreatorID();

  if (!(fe_global.flags & FF_WANTSTORE)) return;

  switch (PREFS_VER) {
  /*case 1:
    WritePrefs_one(cid);
    break;*/

  case 2:
    WritePrefs_two(cid);
    break;

  default:
  }
  WriteOptions(cid);
}

/* --- in-memory read/write callbacks for (de)serialisation --- */

typedef struct _im_rwctx {
    unsigned char *ptr;
    size_t sz, pos;             /* sz = allocated, pos = curr. position */
} im_rwctx;

static int im_readcallback(void *ctx, void *buf, int len) PUZZ_SECT;
static int im_readcallback(void *ctx, void *buf, int len)
{
    im_rwctx *rwctx = (im_rwctx *)ctx;
    size_t spc = rwctx->sz-rwctx->pos;
    int ret = 1;

    /* moves both ptr and pos to suit 'next read position' */

    if (len > spc) {
        len = spc;
        ret = 0;
    }
    memcpy(buf, rwctx->ptr, len);
    rwctx->pos += len;
    rwctx->ptr += len;
    return ret;
}

char *midend_im_deserialise(midend *me, unsigned char *ptr, size_t len)
{
    im_rwctx rwctx;

    rwctx.ptr = ptr;
    rwctx.sz = len;
    rwctx.pos = 0;

    return midend_deserialise(me, im_readcallback, &rwctx);
}

static void im_writecallback(void *ctx, void *buf, int len) PUZZ_SECT;
static void im_writecallback(void *ctx, void *buf, int len)
{
     im_rwctx *rwctx = (im_rwctx *)ctx;
     size_t spc = rwctx->sz-rwctx->pos;

     /* leaves ptr where it is (may realloc it), moves pos so that
      * ptr+pos is 'next write position'. */

     if (len > spc) {
         rwctx->sz = max((rwctx->sz+1)*2, rwctx->sz+len+1);
         rwctx->ptr = sresize(rwctx->ptr, rwctx->sz, unsigned char);
     }
     memcpy(rwctx->ptr+rwctx->pos, buf, len);
     rwctx->pos += len;
}

/* will allocate *ptr_r. */
void midend_im_serialise(midend *me, unsigned char **ptr_r, size_t *len_r)
{
    im_rwctx rwctx;

    rwctx.ptr = NULL; /* or could pre-alloc a small amount */
    rwctx.sz = 0;
    rwctx.pos = 0;

    midend_serialise(me, im_writecallback, &rwctx);

    *ptr_r = rwctx.ptr;
    /* return the length written, not whatever we happened to allocate. */
    *len_r = rwctx.pos;
}

/* vim: set shiftwidth=4 tabstop=8: */
